use std::fmt::Formatter;
use std::ops::Deref;
use std::str::FromStr;
use std::{env, fmt};

use clap::{arg, command, Command, Parser};
use config::{Config, ConfigError, Environment, File};
use rusqlite::{Connection, OpenFlags};
use serde::Deserialize;

#[derive(PartialEq, Debug, Copy, Clone)]
pub enum VcsType {
    Unknown,
    Github,
}

#[derive(PartialEq, Debug)]
pub enum ArgumentParserError {
    UnknownError,
    TypeError,
    NameError,
    UrlError,
    VersionError,
    ContactError,
    CommandMissingError,
}

#[derive(PartialEq, Debug)]
pub enum Action {
    RegisterOrigin {
        vcsType: VcsType,
        name: String,
        repo_name: String,
        owner_name: String,
    },
    RegisterCurrent {
        vcsType: VcsType,
        name: String,
        version: String,
        contact: String,
    },
    Request,
    List,
}

pub struct ArgumentParser {}

impl ArgumentParser {
    pub fn parse(&self, parameters: &Vec<String>) -> Result<Action, ArgumentParserError> {
        let matches = command!()
            .subcommand(Command::new("request"))
            .subcommand(Command::new("list"))
            .subcommand(
                Command::new("register")
                    .subcommand(
                        Command::new("origin").subcommand(
                            Command::new("github")
                                .arg(arg!([name]))
                                .arg(arg!([owner_name]))
                                .arg(arg!([repo_name])),
                        ),
                    )
                    .subcommand(
                        Command::new("current").subcommand(
                            Command::new("github")
                                .arg(arg!([name]))
                                .arg(arg!([version]))
                                .arg(arg!([contact])),
                        ),
                    ),
            )
            .get_matches_from(parameters);

        // TODO: This should be rethought. Right now it covers all, but it is a pain to add more VcsTypes
        match matches.subcommand() {
            Some(("list", sub_matches)) => Ok(Action::List),
            Some(("request", sub_matches)) => Ok(Action::Request),
            Some(("register", sub_matches)) => match sub_matches.subcommand() {
                Some(("current", sub_matches)) => match sub_matches.subcommand() {
                    Some(("github", sub_matches)) => {
                        let name = sub_matches.get_one::<String>("name");
                        let version = sub_matches.get_one::<String>("version");
                        let contact = sub_matches.get_one::<String>("contact");

                        let name = match name {
                            Some(n) => n,
                            None => return Err(ArgumentParserError::NameError),
                        };

                        let version = match version {
                            Some(v) => v,
                            None => return Err(ArgumentParserError::VersionError),
                        };

                        let contact = match contact {
                            Some(c) => c,
                            None => return Err(ArgumentParserError::ContactError),
                        };

                        Ok(Action::RegisterCurrent {
                            version: version.to_string(),
                            name: name.to_string(),
                            contact: contact.to_string(),
                            vcsType: VcsType::Github,
                        })
                    }
                    _ => Err(ArgumentParserError::CommandMissingError),
                },
                Some(("origin", sub_matches)) => match sub_matches.subcommand() {
                    Some(("github", sub_matches)) => {
                        let name = sub_matches.get_one::<String>("name");
                        let owner_name = sub_matches.get_one::<String>("owner_name");
                        let repo_name = sub_matches.get_one::<String>("repo_name");
                        let vcs_type = VcsType::Github;

                        let name = match name {
                            Some(n) => n,
                            None => return Err(ArgumentParserError::NameError),
                        };

                        let owner_name = match owner_name {
                            Some(u) => u,
                            None => return Err(ArgumentParserError::UrlError),
                        };

                        let repo_name = match repo_name {
                            Some(u) => u,
                            None => return Err(ArgumentParserError::UrlError),
                        };
                        Ok(Action::RegisterOrigin {
                            name: name.to_string(),
                            vcsType: vcs_type,
                            owner_name: owner_name.to_string(),
                            repo_name: repo_name.to_string(),
                        })
                    }
                    _ => Err(ArgumentParserError::CommandMissingError),
                },
                _ => Err(ArgumentParserError::CommandMissingError),
            },
            _ => Err(ArgumentParserError::CommandMissingError),
        }
    }
}

#[derive(Debug, Deserialize)]
pub struct Database {
    pub connection_string: String,
    pub in_memory: bool,
}

#[derive(Debug, Deserialize)]
pub struct Email {
    pub from: String,
    pub smtp_username: String,
    pub smtp_password: String,
    pub smtp_relay: String,
}

#[derive(Debug, Deserialize)]
pub struct Settings {
    pub email: Email,
    pub database: Database,
    pub github_api_override: Option<String>, // This is a simple short termed fix that is likely to be deprecated when we get another provider (this is not going a way, but it might be structured differently). This has been made so i can find a way to inject the httpmock server url into application
}

impl Settings {
    pub fn new(github_api_override: Option<&str>) -> Result<Self, ConfigError> {
        let run_mode = env::var("RUN_MODE").unwrap_or_else(|_| "development".into());

        let s = Config::builder()
            .add_source(File::with_name("settings"))
            .add_source(File::with_name(&format!("settings.{}", run_mode)).required(false))
            .add_source(File::with_name("local").required(false))
            .add_source(Environment::with_prefix("version_checker"))
            .build()?;

        let mut settings: Settings = s.try_deserialize().unwrap();

        // TODO: there must be a easier way of expressing this in rust :D
        match github_api_override {
            Some(o) => settings.github_api_override = Some(o.to_string()),
            None => settings.github_api_override = None,
        };

        Ok(settings)
    }
}

impl Database {
    pub fn get_connection(&self) -> Connection {
        if self.in_memory {
            Connection::open_with_flags(
                "file::memory:?cache=shared",
                OpenFlags::SQLITE_OPEN_URI | OpenFlags::SQLITE_OPEN_READ_WRITE,
            )
            .unwrap()
        } else {
            Connection::open(&self.connection_string).unwrap()
        }
    }
}

//TODO: add tests for other commands as well
#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_argument_parser_should_parse_register_origin_subcommand() {
        let arguments = vec![
            String::from("release_checker"),
            String::from("register"),
            String::from("origin"),
            String::from("github"),
            String::from("name"),
            String::from("owner_name"),
            String::from("repo_name"),
        ];
        let parser = ArgumentParser {};

        let result = parser.parse(&arguments);
        assert_eq!(
            result,
            Ok(Action::RegisterOrigin {
                name: String::from("name"),
                vcsType: VcsType::Github,
                owner_name: String::from("owner_name"),
                repo_name: String::from("repo_name"),
            })
        )
    }
}
