use rusqlite::Connection;

use crate::console::{Console, Print};
use crate::list_overview::models::ListOverviewRequest;
use crate::list_overview::OverviewHandler;
use configuration::ArgumentParser;
use configuration::Settings;

use crate::register_current::models::CurrentRequest;
use crate::register_current::RegisterCurrentHandler;
use crate::register_origin::models::OriginRequest;
use crate::register_origin::RegisterOriginHandler;
use crate::request::notification::Notification;
use crate::request::RequestHandler;
use crate::routing::Router;

pub mod configuration;
pub mod console;
pub mod list_overview;
pub mod register_current;
pub mod register_origin;
pub mod request;
mod routing;

mod embedded {
    use refinery::embed_migrations;

    embed_migrations!("src/sql_migrations");
}

pub fn migrate(settings: &Settings) {
    let mut connection = settings.database.get_connection();
    embedded::migrations::runner().run(&mut connection).unwrap();
}

pub struct Application<'a> {
    settings: &'a Settings,
    argument_parser: ArgumentParser,
    arguments: Vec<String>,
    register_origin_handler: RegisterOriginHandler<'a>,
    register_current_handler: RegisterCurrentHandler<'a>,
    overview_handler: OverviewHandler<'a>,
    request_handler: RequestHandler<'a>,
}

impl<'a> Application<'a> {
    pub fn new(
        settings: &'a Settings,
        console: &'a dyn Print,
        notification: &'a dyn Notification,
        argument_parser: ArgumentParser,
        arguments: Vec<String>,
    ) -> Application<'a> {
        Application {
            settings,
            argument_parser,
            arguments,

            // TODO: this init should be rethought to be more extensible if possible
            register_origin_handler: register_origin::init(settings),
            register_current_handler: register_current::init(settings),
            overview_handler: list_overview::init(settings, console),
            request_handler: request::init(settings, console, notification),
        }
    }

    pub fn run(&self) {
        migrate(&self.settings); //TODO: this should be run for it self by calling a subcommand on the cli it self

        // TODO: redo this, because this is just me trying out rust :) I can be done a lot easier, without doing the function delegation
        let register_origin = |name, owner_name, repo_name, vcs_type| {
            self.register_origin_handler.execute(OriginRequest {
                name,
                owner_name,
                repo_name,
                vcs_type,
            });
        };

        let register_current = |name, version, vcs_type, contact| {
            self.register_current_handler.execute(CurrentRequest {
                name,
                version,
                vcs_type,
                contact,
            })
        };

        let list = || {
            self.overview_handler.execute(ListOverviewRequest {
                count: 100, //TODO: hardcoded for now, but needs to be supplied through a cli parameter
            })
        };
        let request = || {
            self.request_handler.execute();
        };

        let action = self.argument_parser.parse(&self.arguments).unwrap();
        let router = Router::new(register_origin, register_current, list, request);
        router.route(action);
    }
}
