use crate::configuration::{Action, VcsType};

pub struct Router<A, F, L, R>
where
    A: Fn(String, String, String, VcsType),
    F: Fn(String, String, VcsType, String),
    L: Fn(),
    R: Fn(),
{
    register_origin: A,
    register_current: F,
    list: L,
    request: R,
}

impl<A, F, L, R> Router<A, F, L, R>
where
    A: Fn(String, String, String, VcsType),
    F: Fn(String, String, VcsType, String),
    L: Fn(),
    R: Fn(),
{
    pub fn new(register_origin: A, register_current: F, list: L, request: R) -> Router<A, F, L, R> {
        Router {
            register_origin,
            register_current,
            list,
            request,
        }
    }
    pub fn route(&self, action: Action) {
        match action {
            Action::RegisterOrigin {
                name,
                owner_name,
                repo_name,
                vcsType,
            } => (self.register_origin)(name, owner_name, repo_name, vcsType),
            Action::Request => (self.request)(),
            Action::List => (self.list)(),
            Action::RegisterCurrent {
                name,
                version,
                vcsType,
                contact,
            } => (self.register_current)(name, version, vcsType, contact),
            _ => {}
        }
    }
}

#[cfg(test)]
mod tests {
    use std::cell::Cell;

    use crate::configuration::Action;
    use crate::configuration::VcsType::Github;
    use crate::routing::Router;

    #[test]
    fn register_origin_is_called_with_correct_parameters() {
        let add = |name, owner_name, repo_name, vcs_type| {
            assert_eq!("name", name);
            assert_eq!("repo_name", repo_name);
            assert_eq!("owner_name", owner_name);
            assert_eq!(Github, vcs_type);
        };

        let current = |name, version, vcs_type, contact| {};

        let list = || {};
        let request = || {};

        let router = Router::new(add, current, list, request);

        router.route(Action::RegisterOrigin {
            name: String::from("name"),
            repo_name: String::from("repo_name"),
            owner_name: String::from("owner_name"),
            vcsType: Github,
        });
    }

    #[test]
    fn register_current_is_called_with_correct_parameters() {
        let add = |name, owner_name, repo_name, vcs_type| {};

        let current = |name, version, vcs_type, contact| {
            assert_eq!("name", name);
            assert_eq!("version", version);
            assert_eq!(Github, vcs_type);
            assert_eq!("contact", contact);
        };

        let list = || {};
        let request = || {};

        let router = Router::new(add, current, list, request);

        router.route(Action::RegisterCurrent {
            name: String::from("name"),
            version: String::from("version"),
            vcsType: Github,
            contact: String::from("contact"),
        });
    }

    #[test]
    fn register_current_is_called_action() {
        let is_add_called = Cell::new(false);
        let is_current_called = Cell::new(false);
        let is_list_called = Cell::new(false);
        let is_request_called = Cell::new(false);

        let add = |name, owner_name, repo_name, vcs_type| {
            is_add_called.set(true);
        };

        let current = |name, version, vcs_type, contact| is_current_called.set(true);

        let list = || {
            is_list_called.set(true);
        };
        let request = || {
            is_request_called.set(true);
        };

        let router = Router::new(add, current, list, request);

        router.route(Action::RegisterCurrent {
            name: String::from("name"),
            version: String::from("version"),
            vcsType: Github,
            contact: String::from("contact"),
        });

        assert!(!is_add_called.get());
        assert!(!is_list_called.get());
        assert!(!is_request_called.get());
        assert!(is_current_called.get());
    }

    #[test]
    fn register_origin_is_called_action() {
        let is_add_called = Cell::new(false);
        let is_current_called = Cell::new(false);
        let is_list_called = Cell::new(false);
        let is_request_called = Cell::new(false);

        let add = |name, owner_name, repo_name, vcs_type| {
            is_add_called.set(true);
        };

        let current = |name, version, vcs_type, contact| is_current_called.set(true);

        let list = || {
            is_list_called.set(true);
        };
        let request = || {
            is_request_called.set(true);
        };

        let router = Router::new(add, current, list, request);

        router.route(Action::RegisterOrigin {
            name: String::from("name"),
            owner_name: String::from("owner_name"),
            repo_name: String::from("repo_name"),
            vcsType: Github,
        });

        assert!(is_add_called.get());
        assert!(!is_list_called.get());
        assert!(!is_request_called.get());
        assert!(!is_current_called.get());
    }
}
